@@ -27,6 +27,7 @@
 #include <QStringList>
 #include <QScopedPointer>
 #include <QRect>
+#include <QCryptographicHash>
 
 using namespace KScreen;
 
@@ -36,6 +37,7 @@ class Q_DECL_HIDDEN Output::Private
     Private():
         id(0),
         type(Unknown),
+        replicationSource(0),
         rotation(None),
         scale(1.0),
         connected(false),
@@ -50,6 +52,7 @@ class Q_DECL_HIDDEN Output::Private
         type(other.type),
         icon(other.icon),
         clones(other.clones),
+        replicationSource(other.replicationSource),
         currentMode(other.currentMode),
         preferredMode(other.preferredMode),
         preferredModes(other.preferredModes),
@@ -80,6 +83,7 @@ class Q_DECL_HIDDEN Output::Private
     QString icon;
     ModeList modeList;
     QList<int> clones;
+    int replicationSource;
     QString currentMode;
     QString preferredMode;
     QStringList preferredModes;
@@ -218,6 +222,16 @@ QString Output::hash() const
     return name();
 }
 
+QString Output::hashMd5() const
+{
+    if (edid() && edid()->isValid()) {
+        return edid()->hash();
+    }
+    const auto hash = QCryptographicHash::hash(name().toLatin1(),
+                                               QCryptographicHash::Md5);
+    return QString::fromLatin1(hash.toHex());
+}
+
 Output::Type Output::type() const
 {
     return d->type;
@@ -473,6 +487,22 @@ void Output::setClones(const QList<int> &outputlist)
     Q_EMIT clonesChanged();
 }
 
+int Output::replicationSource() const
+{
+    return d->replicationSource;
+}
+
+void Output::setReplicationSource(int source)
+{
+    if (d->replicationSource == source) {
+        return;
+    }
+
+    d->replicationSource = source;
+
+    Q_EMIT replicationSourceChanged();
+}
+
 void Output::setEdid(const QByteArray& rawData)
 {
     Q_ASSERT(d->edid.isNull());
@@ -507,17 +537,35 @@ void KScreen::Output::setFollowPreferredMode(bool follow)
     }
 }
 
+bool Output::isPositionable() const
+{
+    return isConnected() && isEnabled() && !replicationSource();
+}
+
+QSize Output::enforcedModeSize() const
+{
+    if (const auto mode = currentMode()) {
+        return mode->size();
+    } else if (const auto mode = preferredMode()) {
+        return mode->size();
+    } else if (d->modeList.count() > 0) {
+        return d->modeList.first()->size();
+    }
+    return QSize();
+}
+
 QRect Output::geometry() const
 {
-    if (!currentMode()) {
+    QSize size = enforcedModeSize();
+    if (!size.isValid()) {
         return QRect();
     }
+    size = size / d->scale;
 
     // We can't use QRect(d->pos, d->size), because d->size does not reflect the
     // actual rotation() set by caller, it's only updated when we get update from
     // KScreen, but not when user changes mode or rotation manually
 
-    QSize size = currentMode()->size() / d->scale;
     if (!isHorizontal()) {
         size = size.transposed();
     }
@@ -579,6 +627,10 @@ void Output::apply(const OutputPtr& other)
         changes << &Output::clonesChanged;
         setClones(other->d->clones);;
     }
+    if (d->replicationSource != other->d->replicationSource) {
+        changes << &Output::replicationSourceChanged;
+        setReplicationSource(other->d->replicationSource);;
+    }
     if (!d->compareModeList(d->modeList, other->d->modeList)) {
         changes << &Output::outputChanged;
         changes << &Output::modesChanged;
@@ -616,6 +668,7 @@ QDebug operator<<(QDebug dbg, const KScreen::OutputPtr &output)
                                   << "pos:" << output->pos() << "res:" << output->size()
                                   << "modeId:" << output->currentModeId()
                                   << "scale:" << output->scale()
+                                  << "clone:" << (output->clones().isEmpty() ? "no" : "yes")
                                   << "followPreferredMode:" << output->followPreferredMode()
                                   << ")";
     } else {
